// MyExt.cpp

#include "dbgexts.h"

HRESULT CALLBACK 
helloworld(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	UNREFERENCED_PARAMETER(args);

	PDEBUG_CONTROL pDebugControl;
	if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Hello World!\n");
		pDebugControl->Release();
	}
	return S_OK;
}

/*
0:000> x test01!g_ptr
012f3370 Test01!g_ptr = 0x012f20e4

0:000> !gptr
012f3370 test01!g_ptr = 0x012f20e4
<string>
*/
HRESULT CALLBACK 
gptr(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	UNREFERENCED_PARAMETER(args);

	IDebugSymbols* pDebugSymbols;
	if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugSymbols), (void **)&pDebugSymbols)))
	{	// Resolve the symbol
		ULONG64 ulAddress = 0;
		if (SUCCEEDED(pDebugSymbols->GetOffsetByName("test01!g_ptr", &ulAddress)))
		{
			IDebugDataSpaces* pDebugDataSpaces;
			if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugDataSpaces), (void **)&pDebugDataSpaces)))
			{	// Read the value of the pointer from the target address space
				ULONG64 ulPtr = 0;
				if (SUCCEEDED(pDebugDataSpaces->ReadPointersVirtual(1, ulAddress, &ulPtr)))
				{
					PDEBUG_CONTROL pDebugControl;
					if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
					{	// Output the values
						pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "%p test01!g_ptr = 0x%p\n", ulAddress, ulPtr);
						pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "%ma\n", ulPtr);
						pDebugControl->Release();
					}
				}
				pDebugDataSpaces->Release();
			}
			pDebugSymbols->Release();
		}
	}
	return S_OK;
}

#define MAX_STACK_FRAMES 100
/*
// Win7 x86
0:000> knL4
 # ChildEBP RetAddr  
00 001bf9dc 76fd48b4 ntdll!KiFastSystemCallRet
01 001bf9e0 752c1876 ntdll!NtDelayExecution+0xc
02 001bfa48 752c1818 KERNELBASE!SleepEx+0x65
03 001bfa58 012f1015 KERNELBASE!Sleep+0xf
0:000> dps @$csp
<snip>
001bfa4c  752c1818 KERNELBASE!Sleep+0xf
001bfa50  00002710
<snip>

// Win7 x64
0:000> knL2
 # Child-SP          RetAddr           Call Site
00 00000000`001cfc08 000007fe`fd9b1203 ntdll!NtDelayExecution+0xa
01 00000000`001cfc10 00000001`3fda101d KERNELBASE!SleepEx+0xab
0:000> r @rsi
rsi=0000000000002710
*/
HRESULT CALLBACK 
sleepy(PDEBUG_CLIENT4 Client, PCSTR args)
{
	UNREFERENCED_PARAMETER(args);
	BOOL bFound = FALSE;

	IDebugControl* pDebugControl;
	if (SUCCEEDED(Client->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		IDebugSymbols* pDebugSymbols;
		if (SUCCEEDED(Client->QueryInterface(__uuidof(IDebugSymbols), (void **)&pDebugSymbols)))
		{
			DEBUG_STACK_FRAME* pDebugStackFrame = (DEBUG_STACK_FRAME*)malloc(sizeof(DEBUG_STACK_FRAME) * MAX_STACK_FRAMES);
			if (pDebugStackFrame != NULL)
			{	// Get the Stack Frames
				memset(pDebugStackFrame, 0, (sizeof(DEBUG_STACK_FRAME) * MAX_STACK_FRAMES));
				ULONG Frames = 0;
				if (SUCCEEDED(pDebugControl->GetStackTrace(0, 0, 0, pDebugStackFrame, MAX_STACK_FRAMES, &Frames)) && (Frames > 0))
				{
					ULONG ProcessorType = 0;
					ULONG SymSize = 0;
					char SymName[4096];
					memset(SymName, 0, 4096);
					ULONG64 Displacement = 0;
					if (SUCCEEDED(pDebugControl->GetEffectiveProcessorType(&ProcessorType)))
					{	
						for (ULONG n=0; n<Frames; n++)
						{	// Use the Effective Processor Type and the contents of the frame to determine existance
							if (SUCCEEDED(pDebugSymbols->GetNameByOffset(pDebugStackFrame[n].InstructionOffset, SymName, 4096, &SymSize, &Displacement)) && (SymSize > 0))
							{
								if ((ProcessorType == IMAGE_FILE_MACHINE_I386) && (_stricmp(SymName, "KERNELBASE!Sleep") == 0) && (Displacement == 0xF))
								{	// Win7 x86; KERNELBASE!Sleep+0xF is usually in frame 3
									IDebugDataSpaces* pDebugDataSpaces;
									if (SUCCEEDED(Client->QueryInterface(__uuidof(IDebugDataSpaces), (void **)&pDebugDataSpaces)))
									{	// The value is pushed immediately prior to KERNELBASE!Sleep+0xF
										DWORD dwMilliseconds = 0;
										if (SUCCEEDED(pDebugDataSpaces->ReadVirtual(pDebugStackFrame[n].StackOffset, &dwMilliseconds, sizeof(dwMilliseconds), NULL)))
										{
											pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Sleeping for %ld msec\n", dwMilliseconds);
											bFound = TRUE;
										}
										pDebugDataSpaces->Release();
									}
									if (bFound) break;
								}
								else if ((ProcessorType == IMAGE_FILE_MACHINE_AMD64) && (_stricmp(SymName, "KERNELBASE!SleepEx") == 0) && (Displacement == 0xAB))
								{	// Win7 x64; KERNELBASE!SleepEx+0xAB is usually in frame 1
									IDebugRegisters* pDebugRegisters;
									if (SUCCEEDED(Client->QueryInterface(__uuidof(IDebugRegisters), (void **)&pDebugRegisters)))
									{	// The value is in the 'rsi' register
										ULONG rsiIndex = 0;
										if (SUCCEEDED(pDebugRegisters->GetIndexByName("rsi", &rsiIndex)))
										{
											DEBUG_VALUE debugValue;
											if (SUCCEEDED(pDebugRegisters->GetValue(rsiIndex, &debugValue)) && (debugValue.Type == DEBUG_VALUE_INT64))
											{	// Truncate to 32bits for display
												pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Sleeping for %ld msec\n", debugValue.I32);
												bFound = TRUE;
											}
										}
										pDebugRegisters->Release();
									}
									if (bFound) break;
								}
							}
						}
					}
				}
				free(pDebugStackFrame);
			}
			pDebugSymbols->Release();
		}
		if (!bFound)
			pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Unable to determine if Sleep is present\n");
		pDebugControl->Release();
	}
	return S_OK;
}
